use anyhow::Result;
use cgmath::{Deg, Point2};
use winit::{
	event::{ElementState, Event, KeyboardInput, VirtualKeyCode, WindowEvent},
	event_loop::{ControlFlow, EventLoop},
	window::WindowBuilder,
};

mod doodad;
mod inputstatus;
mod physics;
mod render;
mod ship;
mod system;

use crate::{
	doodad::Doodad,
	inputstatus::InputStatus,
	physics::Pfloat,
	render::GPUState,
	ship::{Ship, ShipKind},
	system::System,
};

struct Camera {
	pos: Point2<Pfloat>,
}

trait Spriteable {
	fn sprite(&self, camera: &Camera) -> Sprite;
}

struct Sprite {
	// Name of the sprite to draw
	name: String,

	// This object's position, in world coordinates.
	position: Point2<Pfloat>,

	// This sprite's rotation
	// (relative to north, measured ccw)
	angle: Deg<Pfloat>,

	// The camera we want to draw this sprite from, in world coordinates
	camera: Point2<Pfloat>,
}

struct Game {
	input: InputStatus,
	last_update: Instant,
	player: Ship,
	system: System,
	camera: Camera,
}

impl Game {
	fn new() -> Self {
		Game {
			last_update: Instant::now(),
			input: InputStatus::new(),
			player: Ship::new(ShipKind::Gypsum, (0.0, 0.0).into()),
			camera: Camera {
				pos: (0.0, 0.0).into(),
			},
			system: System::new(),
		}
	}

	fn process_input(&mut self, gpu: &ElementState, key: &VirtualKeyCode) {
		self.input.process(gpu, key)
	}

	fn update(&mut self) {
		let t: Pfloat = self.last_update.elapsed().as_secs_f32();
		println!("{:.02}", 1.0 / t);

		if self.input.key_thrust {
			self.player.body.thrust(50.0 * t);
		}

		if self.input.key_right {
			self.player.body.rot(Deg { 0: 15.0 } * t);
		}

		if self.input.key_left {
			self.player.body.rot(Deg { 0: -15.0 } * t);
		}

		self.player.body.tick(t);
		self.camera.pos = self.player.body.pos;

		self.last_update = Instant::now();
	}

	fn sprites(&self) -> Vec<Sprite> {
		let mut sprites: Vec<Sprite> = Vec::new();

		sprites.append(&mut self.system.sprites(&self.camera));
		sprites.push(self.player.sprite(&self.camera));

		return sprites;
	}
}

use std::time::Instant;

pub async fn run() -> Result<()> {
	env_logger::init();
	let event_loop = EventLoop::new();
	let window = WindowBuilder::new().build(&event_loop).unwrap();

	let mut gpu = GPUState::new(window).await?;
	let mut game = Game::new();

	event_loop.run(move |event, _, control_flow| {
		match event {
			Event::RedrawRequested(window_id) if window_id == gpu.window().id() => {
				gpu.update();
				game.update();
				match gpu.render(&game.sprites()) {
					Ok(_) => {}
					// Reconfigure the surface if lost
					Err(wgpu::SurfaceError::Lost) => gpu.resize(gpu.size),
					// The system is out of memory, we should probably quit
					Err(wgpu::SurfaceError::OutOfMemory) => *control_flow = ControlFlow::Exit,
					// All other errors (Outdated, Timeout) should be resolved by the next frame
					Err(e) => eprintln!("{:?}", e),
				}
			}

			Event::MainEventsCleared => {
				// RedrawRequested will only trigger once unless we manually
				// request it.
				gpu.window().request_redraw();
			}

			Event::WindowEvent {
				ref event,
				window_id,
			} if window_id == gpu.window.id() => {
				match event {
					WindowEvent::CloseRequested => *control_flow = ControlFlow::Exit,
					WindowEvent::KeyboardInput {
						input:
							KeyboardInput {
								state,
								virtual_keycode: Some(key),
								..
							},
						..
					} => game.process_input(state, key),
					WindowEvent::Resized(physical_size) => {
						gpu.resize(*physical_size);
					}
					WindowEvent::ScaleFactorChanged { new_inner_size, .. } => {
						// new_inner_size is &&mut so we have to dereference it twice
						gpu.resize(**new_inner_size);
					}
					_ => {}
				}
			}
			_ => {}
		}
	});
}

fn main() -> Result<()> {
	pollster::block_on(run())?;
	return Ok(());
}
